@@ -1,5 +1,4 @@
-/* $Id: ftp.c,v 1.37 2004/05/20 11:10:52 mhe Exp $
- *
+/*
  * ftp.c -- low(er) level FTP stuff
  *
  * Yet Another FTP Client
@@ -17,16 +16,14 @@
 #include "xmalloc.h"
 #include "strq.h"
 #include "gvars.h"
+#ifdef HAVE_LIBSSH
 #include "ssh_cmd.h"
-#include "sftp-common.h"
+#endif
 #include "args.h"
 
 /* in cmd.c */
 void exit_yafc(void);
 
-/* in get.c */
-char *make_unique_filename(const char *path);
-
 /* in tag.c */
 void save_taglist(const char *alt_filename);
 
@@ -37,9 +34,8 @@ Ftp *ftp = 0;
 
 Ftp *ftp_create(void)
 {
-    Ftp *ftp;
-
-    ftp = (Ftp *)xmalloc(sizeof(Ftp));
+    Ftp* ftp = xmalloc(sizeof(Ftp));
+    memset(ftp, 0, sizeof(Ftp));
 
     ftp->verbosity = vbCommand;
     ftp->tmp_verbosity = vbUnset;
@@ -55,8 +51,10 @@ Ftp *ftp_create(void)
 /*  ftp->out_buffer = 0;*/
 #endif
     ftp->LIST_type = ltUnknown;
-    ftp->ssh_id = 1;
-    ftp->ssh_args = args_create();
+#ifdef HAVE_LIBSSH
+    ftp->session = NULL;
+    ftp->sftp_session = NULL;
+#endif
 
     return ftp;
 }
@@ -117,20 +115,20 @@ void ftp_destroy(Ftp *ftp)
 
     list_free(ftp->dirs_to_flush);
     list_free(ftp->cache);
-    ftp->cache = ftp->dirs_to_flush = 0;
+    ftp->cache = ftp->dirs_to_flush = NULL;
     host_destroy(ftp->host);
     sock_destroy(ftp->data);
     sock_destroy(ftp->ctrl);
-    ftp->host = 0;
-    ftp->data = ftp->ctrl = 0;
+    ftp->host = NULL;
+    ftp->data = ftp->ctrl = NULL;
     url_destroy(ftp->url);
-    ftp->url = 0;
+    ftp->url = NULL;
     free(ftp->homedir);
     free(ftp->curdir);
     free(ftp->prevdir);
     list_free(ftp->taglist);
-    args_destroy(ftp->ssh_args);
-    ftp->ssh_args = 0;
+    free(ftp->ti.remote_name);
+    free(ftp->ti.local_name);
 
 #ifdef SECFTP
     sec_end();
@@ -173,24 +171,27 @@ static int proxy_type(url_t *url)
 void ftp_reset_vars(void)
 {
     sock_destroy(ftp->data);
-    ftp->data = 0;
+    ftp->data = NULL;
 
     sock_destroy(ftp->ctrl);
-    ftp->ctrl = 0;
+    ftp->ctrl = NULL;
 
     host_destroy(ftp->host);
-    ftp->host = 0;
+    ftp->host = NULL;
 
-    if(ftp->ssh_pid) {
-        ftp->ssh_pid = 0;
-        close(ftp->ssh_in);
-        close(ftp->ssh_out);
-        ftp->ssh_id = 0;
-    }
-    args_clear(ftp->ssh_args);
+#ifdef HAVE_LIBSSH
+		if (ftp->session)
+		{
+			sftp_free(ftp->sftp_session);
+			ssh_disconnect(ftp->session);
+			ssh_free(ftp->session);
+			ftp->sftp_session = NULL;
+			ftp->session = NULL;
+		}
+#endif
 
     url_destroy(ftp->url);
-    ftp->url = 0;
+    ftp->url = NULL;
 
     ftp->connected = false;
     ftp->loggedin = false;
@@ -203,11 +204,8 @@ void ftp_reset_vars(void)
     ftp->has_site_idle_command = true;
     ftp->has_mlsd_command = true;
 
-    list_free(ftp->dirs_to_flush);
-    ftp->dirs_to_flush = list_new((listfunc)free);
-
-    list_free(ftp->cache);
-    ftp->cache = list_new((listfunc)rdir_destroy);
+    list_clear(ftp->dirs_to_flush);
+    list_clear(ftp->cache);
 
     /* don't assume server is in ascii mode initially even if RFC says so */
     ftp->prev_type = '?';
@@ -258,7 +256,7 @@ static sigjmp_buf open_timeout_jmp;
 static jmp_buf open_timeout_jmp;
 #endif
 
-static RETSIGTYPE ftp_open_handler(int signum)
+static void ftp_open_handler(int signum)
 {
     ftp_longjmp(open_timeout_jmp, 1);
 }
@@ -305,13 +303,23 @@ int ftp_open_url(url_t *urlp, bool reset_vars)
 
     ftp->host = host_create(use_proxy ? gvProxyUrl : urlp);
 
-    if(host_lookup(ftp->host) != 0) {
-        herror(ftp->host->hostname);
+    if (!urlp->protocol || strcmp(urlp->protocol, "ssh") != 0) {
+      // We only need to do a host lookup if we're not using sftp.
+      if(!host_lookup(ftp->host)) {
+        fprintf(stderr, "Host look up for '%s' failed: %s\n",
+            host_getname(ftp->host), host_geterror(ftp->host));
         alarm(0);
         ftp_set_signal(SIGALRM, SIG_IGN);
         return -1;
+      }
     }
+    /* keep the value in urlp->port
     urlp->port = ntohs(ftp->host->port);
+    and set it to 21 if it is -1 */
+    if(urlp->port == -1) {
+	    urlp->port = 21;
+    }
+
 
     fprintf(stderr, "\r               ");
     i = strlen(use_proxy ? gvProxyUrl->hostname : urlp->hostname);
@@ -320,11 +328,13 @@ int ftp_open_url(url_t *urlp, bool reset_vars)
     fprintf(stderr, "\r");
     ftp_trace("\n");
 
+#ifdef HAVE_LIBSSH
     if(urlp->protocol && strcmp(urlp->protocol, "ssh") == 0) {
         int ret = ssh_open_url(urlp);
         alarm(0);
         return ret;
     }
+#endif
 
     if(urlp->protocol && strcmp(urlp->protocol, "ftp") != 0) {
         ftp_err(_("Sorry, don't know how to handle your '%s' protocol\n"
@@ -334,22 +344,31 @@ int ftp_open_url(url_t *urlp, bool reset_vars)
     }
 
     if(use_proxy) {
-        ftp_err(_("Connecting to proxy %s (%s) at port %d...\n"),
-                ftp->host->ohostname, ftp->host->ipnum,
-                urlp->port);
+        ftp_err(_("Connecting to proxy %s at port %d...\n"),
+                host_getoname(ftp->host), urlp->port);
     } else {
-        ftp_err(_("Connecting to %s (%s) at port %d...\n"),
-                ftp->host->ohostname, ftp->host->ipnum,
-                urlp->port);
+        ftp_err(_("Connecting to %s at port %d...\n"),
+                host_getoname(ftp->host), urlp->port);
     }
 
     ftp->ctrl = sock_create();
-    if(sock_connect_host(ftp->ctrl, ftp->host) == -1) {
+    if (!ftp->ctrl) {
+        ftp_err(_("Unable to create socket.\n"));
+        alarm(0);
+        ftp_set_signal(SIGALRM, SIG_IGN);
+        return -1;
+    }
+
+    if(!sock_connect_host(ftp->ctrl, ftp->host)) {
         alarm(0);
         ftp_set_signal(SIGALRM, SIG_IGN);
         return -1;
     }
     sock_lowdelay(ftp->ctrl);
+    char* ip = host_getip(ftp->host);
+    ftp_err(_("Connected to %s ([%s]:%d).\n"),
+        host_getoname(ftp->host), ip, urlp->port);
+    free(ip);
 
     /* read startup message from server */
     ftp_set_tmp_verbosity(vbCommand);
@@ -368,17 +387,19 @@ int ftp_open_url(url_t *urlp, bool reset_vars)
 
     if(ftp->connected) {
         void (*tracefunq)(const char *fmt, ...);
-        unsigned char *a;
 
         url_destroy(ftp->url);
         ftp->url = url_clone(urlp);
 
         tracefunq = (ftp->verbosity == vbDebug ? ftp_err : ftp_trace);
 
-        a = (unsigned char *)&ftp->ctrl->remote_addr.sin_addr;
-        tracefunq("remote address: %d.%d.%d.%d\n", a[0], a[1], a[2], a[3]);
-        a = (unsigned char *)&ftp->ctrl->local_addr.sin_addr;
-        tracefunq("local address: %d.%d.%d.%d\n", a[0], a[1], a[2], a[3]);
+        char* remote_addr = printable_address(sock_remote_addr(ftp->ctrl)),
+            *local_addr = printable_address(sock_local_addr(ftp->ctrl));
+        tracefunq("remote address: %s\n", remote_addr);
+        tracefunq("local address: %s\n", local_addr);
+        free(remote_addr);
+        free(local_addr);
+
         return 0;
     } else {
         ftp_close();
@@ -386,13 +407,6 @@ int ftp_open_url(url_t *urlp, bool reset_vars)
     }
 }
 
-int ftp_open(const char *host)
-{
-    url_t *url;
-    url = url_init(host);
-    return ftp_open_url(ftp->url, true);
-}
-
 int ftp_reopen(void)
 {
     if(ftp && ftp->url) {
@@ -418,8 +432,6 @@ int ftp_reopen(void)
  */
 static int ftp_gets(void)
 {
-    int c, i=0;
-
     ftp->reply[0] = 0;
 
     if(!sock_connected(ftp->ctrl)) {
@@ -427,14 +439,29 @@ static int ftp_gets(void)
         return -1;
     }
 
-    while(true) {
-        c = sock_get(ftp->ctrl);
-        if(c == EOF) {
+    size_t i = 0;
+    bool next = true;
+    int nc = 0;
+    while(true)
+    {
+        int c = 0;
+        if (next)
+          c = sock_get(ftp->ctrl);
+        else
+        {
+          c = nc;
+          next = true;
+        }
+
+        if (c == EOF)
+        {
             ftp_err(_("Server has closed control connection\n"));
             ftp_close();
             return -1;
         }
-        else if(c == 255/*IAC*/) {   /* handle telnet commands */
+        else if(c == 255/*IAC*/)
+        {
+            /* handle telnet commands */
             switch(c = sock_get(ftp->ctrl)) {
               case 251/*WILL*/:
               case 252/*WONT*/:
@@ -453,21 +480,21 @@ static int ftp_gets(void)
             }
             continue;
         }
-        else if(c == '\r') {
-            c = sock_get(ftp->ctrl);
-            if(c == 0)
-                c = '\r';
-            else if(c == '\n') {
+        else if (c == '\r')
+        {
+            nc = sock_get(ftp->ctrl);
+            if (nc == 0)
+              ; // do nothing
+            else if (nc == '\n') {
                 ftp->reply[i++] = (char)c;
                 break;
-            } else if(c == EOF)
-                /* do nothing */ ;
-            else { /* telnet protocol violation, hmpf... */
-                sock_unget(ftp->ctrl, c);
-                continue;
-            }
+            } else if(nc == EOF)
+                c = EOF ;
+            else
+                /* telnet protocol violation, hmpf... */
+                next = false;
         }
-        else if(c == '\n')
+        else if (c == '\n')
             break;
         if(i < MAXREPLY)
             ftp->reply[i++] = (char)c;
@@ -511,8 +538,10 @@ const char *ftp_getreply(bool withcode)
 {
     char *r = ftp->reply;
 
-    if(ftp->ssh_pid)
-        return fx2txt(ftp->ssh_last_status);
+#ifdef HAVE_LIBSSH
+    if (ftp->session)
+        return ssh_get_error(ftp->session);
+#endif
 
     if(withcode)
         return r;
@@ -529,20 +558,20 @@ static void ftp_print_reply(void)
     verbose_t v = ftp_get_verbosity();
 
     ftp_trace("<-- [%s] %s\n",
-              ftp->url ? ftp->url->hostname : ftp->host->hostname,
+              ftp->url ? ftp->url->hostname : host_getname(ftp->host),
               ftp_getreply(true));
 
     if(v >= vbCommand || (ftp->code >= ctTransient && v == vbError)) {
         if(v == vbDebug)
             fprintf(stderr, "<-- [%s] %s\n",
-                    ftp->url ? ftp->url->hostname : ftp->host->hostname,
+                    ftp->url ? ftp->url->hostname : host_getname(ftp->host),
                     ftp_getreply(true));
         else
             fprintf(stderr, "%s\n", ftp_getreply(false));
     }
 }
 
-static RETSIGTYPE reply_ALRM_handler(int signum)
+static void reply_ALRM_handler(int signum)
 {
     ftp_err(_("Tired of waiting for reply, timeout after %u seconds\n"),
             ftp->reply_timeout);
@@ -566,11 +595,12 @@ int ftp_read_reply(void)
     if(ftp->reply_timeout)
         alarm(ftp->reply_timeout);
 
-    clearerr(ftp->ctrl->sin);
+    sock_clearerr_in(ftp->ctrl);
     r = ftp_gets();
     if(!sock_connected(ftp->ctrl)) {
         alarm(0);
         ftp_set_signal(SIGALRM, SIG_DFL);
+        ftp_trace("sock is not connected\n");
         return -1;
     }
 
@@ -604,8 +634,11 @@ static void ftp_print_cmd(const char *cmd, va_list ap)
         if(strncmp(cmd, "PASS", 4) == 0)
             ftp_err("PASS ********");
         else {
+            va_list aq;
+            va_copy(aq, ap);
             vfprintf(stderr, cmd, ap);
-            ftp_vtrace(cmd, ap);
+            ftp_vtrace(cmd, aq);
+            va_end(aq);
         }
         ftp_err("\n");
     } else {
@@ -644,7 +677,7 @@ int ftp_cmd(const char *cmd, ...)
     sock_flush(ftp->ctrl);
     va_end(ap);
 
-    if(ferror(ftp->ctrl->sout)) {
+    if (sock_error_out(ftp->ctrl)) {
         ftp_err(_("error writing command"));
         ftp_err(" (");
         va_start(ap, cmd);
@@ -688,12 +721,17 @@ int ftp_cmd(const char *cmd, ...)
 
 void ftp_quit(void)
 {
-    if(ftp_connected() && !ftp->ssh_pid) {
-        ftp_reply_timeout(10);
-        ftp_set_tmp_verbosity(vbCommand);
-        ftp_cmd("QUIT");
-    }
-    ftp_close();
+#ifdef HAVE_LIBSSH
+	if (ftp_connected() && !ftp->session)
+#else
+	if (ftp_connected())
+#endif
+	{
+		ftp_reply_timeout(10);
+    ftp_set_tmp_verbosity(vbCommand);
+    ftp_cmd("QUIT");
+  }
+  ftp_close();
 }
 
 int ftp_get_verbosity(void)
@@ -725,10 +763,17 @@ int get_username(url_t *url, const char *guessed_username, bool isproxy)
             return -1;
         }
 
-        if(isproxy)
-            asprintf(&prompt, _("Proxy login: "));
-        else if(guessed_username)
-            asprintf(&prompt, _("login (%s): "), guessed_username);
+        if (isproxy)
+            prompt = xstrdup(_("Proxy login: "));
+        else if (guessed_username)
+        {
+            if (asprintf(&prompt, _("login (%s): "), guessed_username) == -1)
+            {
+              fprintf(stderr, _("Failed to allocate memory.\n"));
+              ftp->loggedin = false;
+              return -1;
+            }
+        }
         else
             prompt = xstrdup(_("login (anonymous): "));
         e = ftp->getuser_hook(prompt);
@@ -753,7 +798,7 @@ int get_username(url_t *url, const char *guessed_username, bool isproxy)
     return 0;
 }
 
-static int get_password(url_t *url, const char *anonpass, bool isproxy)
+int get_password(url_t *url, const char *anonpass, bool isproxy)
 {
     if(!url->password) {
         char *e;
@@ -764,7 +809,13 @@ static int get_password(url_t *url, const char *anonpass, bool isproxy)
             e = 0;
             if(ftp->getuser_hook) {
                 if(anonpass && isproxy == false)
-                    asprintf(&prompt, _("password (%s): "), anonpass);
+                {
+                  if (asprintf(&prompt, _("password (%s): "), anonpass) == -1)
+                  {
+                    fprintf(stderr, _("Failed to allocate memory.\n"));
+                    return -1;
+                  }
+                }
                 else
                     prompt = xstrdup(_("password: "));
                 e = ftp->getuser_hook(prompt);
@@ -818,10 +869,8 @@ static const char *secext_name(const char *mech)
 
 static bool mech_unsupported(const char *mech)
 {
-#ifndef HAVE_KRB4
     if(strcasecmp(mech, "KERBEROS_V4") == 0)
         return true;
-#endif
 #ifndef HAVE_KRB5
     if(strcasecmp(mech, "GSSAPI") == 0)
         return true;
@@ -846,9 +895,11 @@ int ftp_login(const char *guessed_username, const char *anonpass)
     if(!ftp->url)
         return -1;
 
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if (ftp->session)
         /* login authentication is performed by the ssh program */
         return 0;
+#endif
 
     ptype = proxy_type(ftp->url);
     if(purl) {
@@ -861,7 +912,7 @@ int ftp_login(const char *guessed_username, const char *anonpass)
     r = get_username(ftp->url, guessed_username, false);
     if(r != 0)
         return r;
-    if(ptype > 1) {
+    if(ptype > 1 && ptype < 7) {
         r = get_username(purl, 0, true);
         if(r != 0)
             return r;
@@ -902,7 +953,7 @@ int ftp_login(const char *guessed_username, const char *anonpass)
                             mech_name);
                     continue;
                 }
-                ret = sec_login(ftp->host->hostname, mech_name);
+                ret = sec_login(host_getname(ftp->host), mech_name);
                 if(ret == -1) {
                     if(ftp->code == ctError
                        && ftp->fullcode != 504 && ftp->fullcode != 534)
@@ -972,6 +1023,9 @@ int ftp_login(const char *guessed_username, const char *anonpass)
                 ftp_cmd("USER %s", ftp->url->username);
         }
         break;
+      case 7:
+        ftp_cmd("USER %s@%s:%i", ftp->url->username, ftp->url->hostname, ftp->url->port);
+        break;
     }
 
     if(ftp->code == ctContinue) {
@@ -1031,6 +1085,7 @@ int ftp_login(const char *guessed_username, const char *anonpass)
                     level_to_name(ftp->data_prot));
         }
 #endif
+
         ftp->homedir = ftp_getcurdir();
         ftp->curdir = xstrdup(ftp->homedir);
         ftp->prevdir = xstrdup(ftp->homedir);
@@ -1051,13 +1106,19 @@ bool ftp_loggedin(void)
 
 bool ftp_connected(void)
 {
-    return (ftp->connected && (sock_connected(ftp->ctrl) || ftp->ssh_pid));
+#ifdef HAVE_LIBSSH
+  return ftp->connected && (sock_connected(ftp->ctrl) || ftp->session);
+#else
+  return ftp->connected && sock_connected(ftp->ctrl);
+#endif
 }
 
 char *ftp_getcurdir(void)
 {
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if (ftp->session)
         return ssh_getcurdir();
+#endif
 
     ftp_set_tmp_verbosity(vbNone);
     ftp_cmd("PWD");
@@ -1097,8 +1158,10 @@ static void ftp_update_curdir(void)
 
 int ftp_chdir(const char *path)
 {
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if (ftp->session)
         return ssh_chdir(path);
+#endif
 
     ftp_set_tmp_verbosity(vbCommand);
     ftp_cmd("CWD %s", path);
@@ -1145,8 +1208,10 @@ int ftp_chdir(const char *path)
 
 int ftp_cdup(void)
 {
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_cdup();
+#endif
 
     ftp_set_tmp_verbosity(vbCommand);
     ftp_cmd("CDUP");
@@ -1161,8 +1226,10 @@ static int ftp_mkdir_verb(const char *path, verbose_t verb)
 {
     char *p;
 
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_mkdir_verb(path, verb);
+#endif
 
     p = xstrdup(path);
     stripslash(p);
@@ -1184,8 +1251,10 @@ int ftp_rmdir(const char *path)
 {
     char *p;
 
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_rmdir(path);
+#endif
 
     p = xstrdup(path);
     stripslash(p);
@@ -1201,8 +1270,10 @@ int ftp_rmdir(const char *path)
 
 int ftp_unlink(const char *path)
 {
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_unlink(path);
+#endif
 
     ftp_cmd("DELE %s", path);
     if(ftp->code == ctComplete) {
@@ -1214,8 +1285,10 @@ int ftp_unlink(const char *path)
 
 int ftp_chmod(const char *path, const char *mode)
 {
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_chmod(path, mode);
+#endif
 
     if(ftp->has_site_chmod_command) {
         ftp_set_tmp_verbosity(vbNone);
@@ -1239,8 +1312,10 @@ void ftp_reply_timeout(unsigned int secs)
 
 int ftp_idle(const char *idletime)
 {
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_idle(idletime);
+#endif
 
     if(!ftp->has_site_idle_command) {
         ftp_err(_("Server doesn't support SITE IDLE\n"));
@@ -1260,8 +1335,10 @@ int ftp_idle(const char *idletime)
 
 int ftp_noop(void)
 {
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_noop();
+#endif
 
     ftp_set_tmp_verbosity(vbCommand);
     ftp_cmd("NOOP");
@@ -1270,8 +1347,10 @@ int ftp_noop(void)
 
 int ftp_help(const char *arg)
 {
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_help(arg);
+#endif
 
     ftp_set_tmp_verbosity(vbCommand);
     if(arg)
@@ -1283,10 +1362,10 @@ int ftp_help(const char *arg)
 
 unsigned long long ftp_filesize(const char *path)
 {
-    unsigned long long ret;
-
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if(ftp->session)
         return ssh_filesize(path);
+#endif
 
     if(!ftp->has_size_command)
         return -1;
@@ -1299,6 +1378,7 @@ unsigned long long ftp_filesize(const char *path)
         return -1;
     }
     if(ftp->code == ctComplete) {
+        unsigned long long ret;
         sscanf(ftp->reply, "%*s %llu", &ret);
         return ret;
     }
@@ -1310,40 +1390,26 @@ rdirectory *ftp_read_directory(const char *path)
     FILE *fp = 0;
     rdirectory *rdir;
     bool is_curdir = false;
-    char *tmpfilename, *e;
     bool _failed = false;
     char *dir;
     bool is_mlsd = false;
 
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if (ftp->session)
         return ssh_read_directory(path);
+#endif
 
     dir = ftp_path_absolute(path);
     stripslash(dir);
 
     is_curdir = (strcmp(dir, ftp->curdir) == 0);
 
-    asprintf(&e, "%s/yafclist.tmp", gvWorkingDirectory);
-    tmpfilename = make_unique_filename(e);
-    free(e);
-    fp = fopen(tmpfilename, "w+");
-
-    if(fp == 0) {
-        /* can't create a tmpfile in ~/.yafc, try in /tmp/ */
-
-        tmpfilename = make_unique_filename("/tmp/yafclist.tmp");
-        fp = fopen(tmpfilename, "w+");
-
-        if(fp == 0) {
-            ftp_err("%s: %s\n", tmpfilename, strerror(errno));
-            free(dir);
-            free(tmpfilename);
-            return 0;
-        }
+    if((fp = tmpfile()) == NULL) {	/* can't create a tmpfile */
+        ftp_err("Unable to create temp file: %s\n", strerror(errno));
+        free(dir);
+        return 0;
     }
 
-    fchmod(fileno(fp), S_IRUSR | S_IWUSR);
-
     /* we do a "CWD" before the listing, because: we want a listing of
      *  the directory contents, not the directory itself, and some
      *  servers misunderstand this. If the target is a link to a
@@ -1393,6 +1459,7 @@ rdirectory *ftp_read_directory(const char *path)
     rdir = rdir_create();
     if(rdir_parse(rdir, fp, dir, is_mlsd) != 0) {
         rdir_destroy(rdir);
+        rdir = NULL;
         goto failed;
     }
 
@@ -1400,17 +1467,15 @@ rdirectory *ftp_read_directory(const char *path)
     ftp_trace("added directory '%s' to cache\n", dir);
     list_additem(ftp->cache, rdir);
     free(dir);
-    unlink(tmpfilename);
-    free(tmpfilename);
+
+    rdir_sort(rdir);
     return rdir;
 
-  failed: /* forgive me father, for I have goto'ed */
-    free(dir);
-    if(fp)
+failed: /* forgive me father, for I have goto'ed */
+    if (fp)
         fclose(fp);
-    unlink(tmpfilename);
-    free(tmpfilename);
-    return 0;
+    free(dir);
+    return NULL;
 }
 
 rdirectory *ftp_get_directory(const char *path)
@@ -1488,7 +1553,7 @@ static bool ftp_path_part_of(const char *a, const char *b)
 int ftp_mkpath(const char *path)
 {
     bool one_created = false;
-    char *p, *orgp, *e = 0;
+    char *p, *orgp, *e = NULL;
 
     if(!path)
         return 0;
@@ -1506,25 +1571,32 @@ int ftp_mkpath(const char *path)
     unquote(p);
 
     if(*p == '/') {
-        e = (char *)xmalloc(1);
+        e = xmalloc(1);
         *e = 0;
     }
 
     while(true) {
-        char *tmp, *foo;
-        tmp = strqsep(&p, '/');
+        char* tmp = strqsep(&p, '/');
         if(!tmp)
             break;
 
-        if(e)
-            asprintf(&foo, "%s/%s", e, tmp);
+        char* foo = NULL;
+        if (e)
+        {
+            if (asprintf(&foo, "%s/%s", e, tmp) == -1)
+            {
+              fprintf(stderr, _("Failed to allocate memory.\n"));
+              free(e);
+              free(orgp);
+              return -1;
+            }
+        }
         else
             foo = xstrdup(tmp);
 
         free(e);
         e = foo;
 
-
         /* check if we already has created this path */
         if(ftp_path_part_of(e, ftp->last_mkpath))
             continue;
@@ -1533,7 +1605,7 @@ int ftp_mkpath(const char *path)
         if(ftp_path_part_of(e, ftp->curdir))
             continue;
 
-        if(strcmp(e, ".") != 0) {
+        if(strncmp(e, ".", 2) != 0) {
             ftp_mkdir_verb(e, vbNone);
             one_created = (ftp->code == ctComplete);
         }
@@ -1554,14 +1626,13 @@ char *ftp_path_absolute(const char *path)
 
 void ftp_flush_reply(void)
 {
-    fd_set ready;
-    struct timeval poll;
-
     if(!ftp_connected())
         return;
 
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if (ftp->session)
         return;
+#endif
 
 /*  ftp_set_signal(SIGINT, SIG_IGN);*/
     fprintf(stderr, "flushing replies...\r");
@@ -1569,11 +1640,7 @@ void ftp_flush_reply(void)
 /*  ftp_reply_timeout(10);*/
 
     while(ftp_connected()) {
-        poll.tv_sec = 1;
-        poll.tv_usec = 0;
-        FD_ZERO(&ready);
-        FD_SET(ftp->ctrl->handle, &ready);
-        if(select(ftp->ctrl->handle+1, &ready, 0, 0, &poll) == 1)
+        if (sock_check_pending(ftp->ctrl, false) == 1)
             ftp_read_reply();
         else
             break;
@@ -1590,8 +1657,10 @@ int ftp_rename(const char *oldname, const char *newname)
     char *on;
     char *nn;
 
-    if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+    if (ftp->session)
         return ssh_rename(oldname, newname);
+#endif
 
     on = xstrdup(oldname);
     stripslash(on);
@@ -1648,7 +1717,7 @@ time_t gmt_mktime(const struct tm *ts)
     const int daysPerMonth[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30,
         31, 30, 31 };
     const int leapDaysPerMonth[] = { 31, 29, 31, 30, 31,
-        30, 31, 31, 30, 31, 30, 31 }; 
+        30, 31, 31, 30, 31, 30, 31 };
     const int *maxDaysInMonth = daysPerMonth;
 
     time_t gmt = -1;
@@ -1707,34 +1776,42 @@ time_t gmt_mktime(const struct tm *ts)
 ** Returns; Either -1 upon failure or the time_t file modification
 **      time upon success.
 */
-time_t ftp_filetime(const char *filename)
+time_t ftp_filetime(const char *filename, bool force)
 {
-    struct tm ts;
+  if (!ftp_connected())
+    return (time_t)-1;
 
-    if(!ftp_connected())
-        return -1;
+  if (!force)
+  {
+    rfile* f = ftp_cache_get_file(filename);
+    if (f && f->mtime != (time_t)-1)
+      return f->mtime;
+  }
 
-    if(ftp->ssh_pid)
-        return ssh_filetime(filename);
+#ifdef HAVE_LIBSSH
+  if (ftp->session)
+    return ssh_filetime(filename);
+#endif
 
-    if(!ftp->has_mdtm_command)
-        return -1;
+  if (!ftp->has_mdtm_command)
+    return (time_t)-1;
 
-    memset(&ts, 0, sizeof(ts));
-    ftp_set_tmp_verbosity(vbNone);
-    ftp_cmd("MDTM %s", filename);
-    if (ftp->fullcode == 202) {
-        ftp->has_mdtm_command = false;
-        return -1;
-    }
-    if (ftp->fullcode != 213)
-        return -1;
-    /* time is Universal Coordinated Time */
-    sscanf(ftp->reply, "%*s %04d%02d%02d%02d%02d%02d", &ts.tm_year,
-           &ts.tm_mon, &ts.tm_mday, &ts.tm_hour, &ts.tm_min, &ts.tm_sec);
-    ts.tm_year -= 1900;
-    ts.tm_mon--;
-    return gmt_mktime(&ts);
+  struct tm ts;
+  memset(&ts, 0, sizeof(ts));
+  ftp_set_tmp_verbosity(vbNone);
+  ftp_cmd("MDTM %s", filename);
+  if (ftp->fullcode == 202) {
+    ftp->has_mdtm_command = false;
+    return (time_t)-1;
+  }
+  if (ftp->fullcode != 213)
+    return (time_t)-1;
+  /* time is Universal Coordinated Time */
+  sscanf(ftp->reply, "%*s %04d%02d%02d%02d%02d%02d", &ts.tm_year,
+         &ts.tm_mon, &ts.tm_mday, &ts.tm_hour, &ts.tm_min, &ts.tm_sec);
+  ts.tm_year -= 1900;
+  ts.tm_mon--;
+  return gmt_mktime(&ts);
 }
 
 int ftp_maybe_isdir(rfile *fp)
@@ -1764,10 +1841,13 @@ int ftp_maybe_isdir(rfile *fp)
 
 void ftp_pwd(void)
 {
-    if(ftp->ssh_pid) {
+#ifdef HAVE_LIBSSH
+    if (ftp->session)
+    {
         ssh_pwd();
         return;
     }
+#endif
 
     ftp_set_tmp_verbosity(vbCommand);
     ftp_cmd("PWD");
@@ -1776,48 +1856,81 @@ void ftp_pwd(void)
 
 char *perm2string(int perm)
 {
-    char *attr = (char *)xmalloc(11);
+  char* attr = xmalloc(11);
+  strncpy(attr, "----------", 11);
 
-    strcpy(attr, "----------");
+  if (S_ISDIR(perm))
+    attr[0] = 'd';
+  else if (S_ISLNK(perm))
+    attr[0] = 'l';
+  else if (S_ISCHR(perm))
+    attr[0] = 'c';
+  else if (S_ISBLK(perm))
+    attr[0] = 'b';
+  else if (S_ISFIFO(perm))
+    attr[0] = 'p';
+  else if (S_ISSOCK(perm))
+    attr[0] = 's';
 
-    if(S_ISDIR(perm))
-        attr[0] = 'd';
-    else if(S_ISLNK(perm))
-        attr[0] = 'l';
-    else if(S_ISCHR(perm))
-        attr[0] = 'c';
-    else if(S_ISBLK(perm))
-        attr[0] = 'b';
-    else if(S_ISFIFO(perm))
-        attr[0] = 'p';
-    else if(S_ISSOCK(perm))
-        attr[0] = 's';
+  if (test(perm, S_IRUSR))
+    attr[1] = 'r';
+  if (test(perm, S_IWUSR))
+    attr[2] = 'w';
+  if (test(perm, S_IXUSR))
+  {
+    if(test(perm, S_ISUID))
+      attr[3] = 's';
+    else
+      attr[3] = 'x';
+  }
+  else if (test(perm, S_ISUID))
+    attr[3] = 'S';
 
-    if(test(perm, S_IRUSR)) attr[1] = 'r';
-    if(test(perm, S_IWUSR)) attr[2] = 'w';
-    if(test(perm, S_IXUSR)) {
-        if(test(perm, S_ISUID)) attr[3] = 's';
-        else attr[3] = 'x';
-    } else if(test(perm, S_ISUID)) attr[3] = 'S';
+  if (test(perm, S_IRGRP))
+    attr[4] = 'r';
+  if (test(perm, S_IWGRP))
+    attr[5] = 'w';
+  if (test(perm, S_IXGRP))
+  {
+    if (test(perm, S_ISGID))
+      attr[6] = 's';
+    else
+      attr[6] = 'x';
+  }
+  else if (test(perm, S_ISGID))
+    attr[6] = 'S';
 
-    if(test(perm, S_IRGRP)) attr[4] = 'r';
-    if(test(perm, S_IWGRP)) attr[5] = 'w';
-    if(test(perm, S_IXGRP)) {
-        if(test(perm, S_ISGID)) attr[6] = 's';
-        else attr[6] = 'x';
-    } else if(test(perm, S_ISGID)) attr[6] = 'S';
+  if (test(perm, S_IROTH))
+    attr[7] = 'r';
+  if (test(perm, S_IWOTH))
+    attr[8] = 'w';
+  if (test(perm, S_IXOTH))
+  {
+    if(test(perm, S_ISVTX))
+      attr[9] = 't';
+    else
+      attr[9] = 'x';
+  }
+  else if (test(perm, S_ISVTX))
+    attr[9] = 'T';
 
-    if(test(perm, S_IROTH)) attr[7] = 'r';
-    if(test(perm, S_IWOTH)) attr[8] = 'w';
-    if(test(perm, S_IXOTH)) {
-        if(test(perm, S_ISVTX)) attr[9] = 't';
-        else attr[9] = 'x';
-    } else if(test(perm, S_ISVTX)) attr[9] = 'T';
-
-    return attr;
+  return attr;
 }
 
 void ftp_get_feat(void)
 {
 /*  ftp_cmd("FEAT");*/
 }
+
+char* ftp_connected_user()
+{
+  if (!ftp_loggedin())
+    return NULL;
+
+#ifdef HAVE_LIBSSH
+  if (ftp->session)
+    return ssh_connected_user();
+#endif
+
+  return ftp->url->username ? xstrdup(ftp->url->username) : NULL;
+}
