@@ -1,5 +1,4 @@
-/* $Id: ftpsend.c,v 1.18 2005/10/05 19:32:47 splicednetworks Exp $
- *
+/*
  * ftpsend.c -- send/receive files and file listings
  *
  * Yet Another FTP Client
@@ -18,48 +17,118 @@
 #include "xmalloc.h"
 #include "ftpsigs.h"
 #include "gvars.h"
+#ifdef HAVE_LIBSSH
 #include "ssh_cmd.h"
-#include "ssh_ftp.h"
+#endif
 
-static int ftp_pasv(unsigned char result[6])
+static bool is_private(struct sockaddr* sa)
 {
-	int pa[6];
-	char *e;
-	int i;
+  if (sa->sa_family == AF_INET)
+  {
+    struct sockaddr_in* in = (struct sockaddr_in*) sa;
+    unsigned char* addr = (unsigned char*) &in->sin_addr.s_addr;
+    return (addr[0] == 10 ||
+        (addr[0] == 172 && addr[1] >= 16 && addr[1] < 32) ||
+        (addr[0] == 192 && addr[1] == 168));
+  }
+  return false;
+}
 
-	if(!ftp->has_pasv_command) {
-		ftp_err(_("Host doesn't support passive mode\n"));
-		return -1;
-	}
-	ftp_set_tmp_verbosity(vbNone);
+static bool is_multicast(struct sockaddr* sa)
+{
+  if (sa->sa_family == AF_INET)
+  {
+    struct sockaddr_in* in = (struct sockaddr_in*) sa;
+    unsigned char* addr = (unsigned char*) &in->sin_addr.s_addr;
+    return (addr[0] >= 224 && addr[0] < 240);
+  }
+  return false;
+}
 
-	/* request passive mode */
-	ftp_cmd("PASV");
+static bool is_loopback(struct sockaddr* sa)
+{
+  if (sa->sa_family == AF_INET)
+  {
+    struct sockaddr_in* in = (struct sockaddr_in*) sa;
+    unsigned char* addr = (unsigned char*) &in->sin_addr.s_addr;
+    return (addr[0] >= 127 && addr[1] == 0 && addr[2] == 0 && addr[3] == 1);
+  }
+  return false;
+}
 
-	if(!ftp_connected())
-		return -1;
 
-	if(ftp->code != ctComplete) {
-		ftp_err(_("Unable to enter passive mode\n"));
-		if(ftp->code == ctError) /* no use try it again */
-			ftp->has_pasv_command = false;
-		return -1;
-	}
+static bool is_reserved(struct sockaddr* sa)
+{
+  if (sa->sa_family == AF_INET)
+  {
+    struct sockaddr_in* in = (struct sockaddr_in*) sa;
+    unsigned char* addr = (unsigned char*) &in->sin_addr.s_addr;
+    return (addr[0] == 0 ||
+        (addr[0] == 127 && !is_loopback(sa)) ||
+        addr[0] >= 240);
+  }
+  return false;
+}
 
-	e = ftp->reply + 4;
-	while(!isdigit((int)*e))
-		e++;
+static bool ftp_pasv(bool ipv6, unsigned char* result, unsigned short* ipv6_port)
+{
+  if (!ftp->has_pasv_command) {
+    ftp_err(_("Host doesn't support passive mode\n"));
+    return false;
+  }
+  ftp_set_tmp_verbosity(vbNone);
 
-	if(sscanf(e, "%d,%d,%d,%d,%d,%d",
-			  &pa[0], &pa[1], &pa[2], &pa[3], &pa[4], &pa[5]) != 6) {
-		ftp_err(_("Error parsing PASV reply: '%s'\n"),
-				ftp_getreply(false));
-		return -1;
-	}
-	for(i=0; i<6; i++)
-		result[i] = (unsigned char)(pa[i] & 0xFF);
+  /* request passive mode */
+  if (!ipv6)
+    ftp_cmd("PASV");
+#ifdef HAVE_IPV6
+  else if (ipv6)
+    ftp_cmd("EPSV");
+#endif
+  else
+    return false;
 
-	return 0;
+  if (!ftp_connected())
+    return false;
+
+  if (ftp->code != ctComplete) {
+    ftp_err(_("Unable to enter passive mode\n"));
+    if(ftp->code == ctError) /* no use try it again */
+      ftp->has_pasv_command = false;
+    return false;
+  }
+
+  const char* e = ftp_getreply(false);
+  while (e && !isdigit((int)*e))
+    e++;
+
+  if (!e)
+  {
+    ftp_err(_("Error parsing EPSV/PASV reply: '%s'\n"), ftp_getreply(false));
+    return false;
+  }
+
+  if (!ipv6)
+  {
+    if (sscanf(e, "%hhu,%hhu,%hhu,%hhu,%hhu,%hhu", &result[0], &result[1],
+          &result[2], &result[3], &result[4], &result[5]) != 6)
+    {
+      ftp_err(_("Error parsing PASV reply: '%s'\n"), ftp_getreply(false));
+      return false;
+    }
+  }
+#ifdef HAVE_IPV6
+  else
+  {
+    if (sscanf(e, "%hu", ipv6_port) != 1)
+    {
+      ftp_err(_("Error parsing EPSV reply: '%s'\n"), ftp_getreply(false));
+      return false;
+    }
+  }
+#endif
+
+  return true;
 }
 
 static bool ftp_is_passive(void)
@@ -71,56 +140,125 @@ static bool ftp_is_passive(void)
 
 static int ftp_init_transfer(void)
 {
-	struct sockaddr_in sa;
-	unsigned char *a, *p;
-	unsigned char pac[6];
+  if (!ftp_connected())
+    return -1;
 
-	if(!ftp_connected())
-		goto err0;
+  if (!sock_dup(ftp->ctrl, &ftp->data))
+    return -1;
 
-	if (!(ftp->data = sock_create())) {
-		goto err0;
-	}
-	sock_copy(ftp->data, ftp->ctrl);
+  if (ftp_is_passive())
+  {
+    ftp_trace("Initializing passive connection.\n");
 
-	if(ftp_is_passive()) {
-		if(ftp_pasv(pac) != 0) {
-			goto err1;
-		}
+    struct sockaddr_storage sa;
+    memcpy(&sa, sock_remote_addr(ftp->ctrl), sizeof(struct sockaddr_storage));
 
-		sock_getsockname(ftp->ctrl, &sa);
-		memcpy(&sa.sin_addr, pac, (size_t)4);
-		memcpy(&sa.sin_port, pac+4, (size_t)2);
-		if(sock_connect_addr(ftp->data, &sa) == -1)
-			goto err1;
-	} else {
-		sock_listen(ftp->data);
+    unsigned char pac[6] = { 0 };
+    unsigned short ipv6_port = { 0 };
+    if (!ftp_pasv(sa.ss_family != AF_INET, pac, &ipv6_port))
+    {
+      ftp_trace("PASV/EPSV failed.\n");
+      sock_destroy(ftp->data);
+      ftp->data = NULL;
+      return -1;
+    }
 
-		a = (unsigned char *)&ftp->data->local_addr.sin_addr;
-		p = (unsigned char *)&ftp->data->local_addr.sin_port;
+    socklen_t len = sizeof(struct sockaddr_in);
+    if (sa.ss_family == AF_INET)
+    {
+      memcpy(&((struct sockaddr_in*)&sa)->sin_addr, pac, (size_t)4);
+      memcpy(&((struct sockaddr_in*)&sa)->sin_port, pac+4, (size_t)2);
+    }
+#ifdef HAVE_IPV6
+    else if (sa.ss_family == AF_INET6)
+    {
+      ((struct sockaddr_in6*)&sa)->sin6_port = htons(ipv6_port);
+      len = sizeof(struct sockaddr_in6);
+    }
+#endif
+    else
+    {
+      ftp_trace("Do not know how to handle family %d.\n", sa.ss_family);
+      sock_destroy(ftp->data);
+      ftp->data = NULL;
+      return -1;
+    }
 
-		ftp_set_tmp_verbosity(vbError);
-		ftp_cmd("PORT %d,%d,%d,%d,%d,%d",
-				a[0], a[1], a[2], a[3], p[0], p[1]);
-		if(ftp->code != ctComplete)
-			goto err1;
-	}
+    struct sockaddr_storage tmp;
+    memcpy(&tmp, sock_remote_addr(ftp->ctrl), sizeof(struct sockaddr_storage));
+    if (is_reserved((struct sockaddr*) &sa) ||
+         is_multicast((struct sockaddr*) &sa)  ||
+         (is_private((struct sockaddr*) &sa) != is_private((struct sockaddr*) &tmp)) ||
+         (is_loopback((struct sockaddr*) &sa) != is_loopback((struct sockaddr*) &tmp)))
+    {
+      // Invalid address returned by PASV. Replace with address from control
+      // socket.
+      ftp_err(_("Address returned by PASV seems to be incorrect.\n"));
+      ((struct sockaddr_in*)&sa)->sin_addr = ((struct sockaddr_in*)&tmp)->sin_addr;
+    }
 
-	sock_throughput(ftp->data);
+    if (!sock_connect_addr(ftp->data, (struct sockaddr*) &sa, len))
+    {
+      ftp_trace("Could not connect to address from PASV/EPSV.\n");
+      perror("connect()");
+      sock_destroy(ftp->data);
+      ftp->data = NULL;
+      return -1;
+    }
+  } else {
+    ftp_trace("Initializing active connection.\n");
 
-	return 0;
+    const struct sockaddr* local = sock_local_addr(ftp->data);
+    sock_listen(ftp->data, local->sa_family);
 
- err1:
-	sock_destroy(ftp->data);
- err0:
-	return -1;
+    if (local->sa_family == AF_INET)
+    {
+      struct sockaddr_in* tmp = (struct sockaddr_in*)local;
+      unsigned char* a = (unsigned char *)&tmp->sin_addr;
+      unsigned char* p = (unsigned char *)&tmp->sin_port;
+
+      ftp_set_tmp_verbosity(vbError);
+      ftp_cmd("PORT %d,%d,%d,%d,%d,%d",
+          a[0], a[1], a[2], a[3], p[0], p[1]);
+    }
+#ifdef HAVE_IPV6
+    else if (local->sa_family == AF_INET6)
+    {
+      char* addr = printable_address(local);
+
+      ftp_set_tmp_verbosity(vbError);
+      ftp_cmd("EPRT |2|%s|%u|", addr, ntohs(((struct sockaddr_in6*)local)->sin6_port));
+      free(addr);
+    }
+#endif
+    else
+    {
+      ftp_trace("Do not know how to handle family %d.\n", local->sa_family);
+      sock_destroy(ftp->data);
+      ftp->data = NULL;
+      return -1;
+    }
+
+    if(ftp->code != ctComplete)
+    {
+      ftp_trace("PORT/EPRT not successful\n");
+      sock_destroy(ftp->data);
+      ftp->data = NULL;
+      return -1;
+    }
+  }
+
+  sock_throughput(ftp->data);
+  return 0;
 }
 
 int ftp_type(transfer_mode_t type)
 {
-	if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+	if (ftp->session)
 		/* FIXME: is this relevant for ssh ? */
 		return 0;
+#endif
 
 	if(type == tmCurrent)
 		return 0;
@@ -138,28 +276,25 @@ static ftp_transfer_func foo_hookf = 0;
 
 /* abort routine originally from Cftp by Dieter Baron
  */
-int ftp_abort(FILE *fp)
+int ftp_abort(Socket* fp)
 {
 	char buf[4096];
-	fd_set ready;
-	struct timeval poll;
 
-	if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+	if(ftp->session)
 		/* FIXME: what? */
 		return 0;
+#endif
 
 	if(!ftp_connected())
 		return -1;
 
 	ftp_set_close_handler();
 
-	poll.tv_sec = poll.tv_usec = 0;
-	FD_ZERO(&ready);
-	FD_SET(ftp->ctrl->handle, &ready);
-	if(select(ftp->ctrl->handle+1, &ready, 0, 0, &poll) == 1) {
+	if (sock_check_pending(fp, false) == 1) {
 		ftp_trace("There is data on the control channel, won't send ABOR\n");
 		/* read remaining bytes from connection */
-		while(fp && fread(buf, 1, 4096, fp) > 0)
+		while(fp && sock_read(fp, buf, sizeof(buf)) > 0)
 			/* LOOP */ ;
 		return 0;
 	}
@@ -183,7 +318,7 @@ int ftp_abort(FILE *fp)
 		ftp_trace("--> [%s] ABOR\n", ftp->url->hostname);
 
     /* read remaining bytes from connection */
-	while(fp && fread(buf, 1, 4096, fp) > 0)
+	while(fp && sock_read(fp, buf, sizeof(buf)) > 0)
 		/* LOOP */ ;
 
 	/* we expect a 426 or 226 reply here... */
@@ -200,23 +335,10 @@ int ftp_abort(FILE *fp)
 	return -1;
 }
 
-static int wait_for_data(FILE *fd, bool wait_for_read)
+static int wait_for_data(Socket* fp, bool wait_for_read)
 {
-	fd_set fds;
-	struct timeval tv;
-	int r;
-
-	/* watch fd to see if it has input */
-	FD_ZERO(&fds);
-	FD_SET(fileno(fd), &fds);
-	/* wait max 1 second */
-	tv.tv_sec = 10;
-	tv.tv_usec = 0;
-
-	if(wait_for_read)
-		r = select(fileno(fd)+1, &fds, 0, 0, &tv);
-	else /* wait for write */
-		r = select(fileno(fd)+1, 0, &fds, 0, &tv);
+  errno = 0;
+	const int r = sock_check_pending(fp, !wait_for_read);
 #if 0
 	if(r < 0 && errno == EINTR && gvSigStopReceived && ftp_sigints() == 0) {
 		gvSigStopReceived = false;
@@ -249,7 +371,7 @@ static int wait_for_input(void)
 {
 	int r;
 	do {
-		r = wait_for_data(ftp->data->sin, true);
+		r = wait_for_data(ftp->data, true);
 		if(r == -1) {
 			if(errno == EINTR)
 				ftp->ti.interrupted = true;
@@ -266,7 +388,7 @@ static int wait_for_output(void)
 {
 	int r;
 	do {
-		r = wait_for_data(ftp->data->sout, false);
+		r = wait_for_data(ftp->data, false);
 		if(r == -1)
 			return -1;
 		if(r == 0 && foo_hookf)
@@ -276,11 +398,9 @@ static int wait_for_output(void)
 	return 0;
 }
 
-static int maybe_abort(FILE *in, FILE *out)
+static int maybe_abort_in(Socket* in, FILE *out)
 {
-	unsigned i;
-
-	i = ftp_sigints();
+	unsigned int i = ftp_sigints();
 	ftp_set_close_handler();
 
 	ftp->ti.finished = true;
@@ -288,8 +408,8 @@ static int maybe_abort(FILE *in, FILE *out)
 	if(ftp->ti.interrupted)
 		i++;
 
-	if(i > 0 || ferror(in) || ferror(out)) {
-		if(ferror(in)) {
+	if(i > 0 || sock_error_in(in) || ferror(out)) {
+		if(sock_error_in(in)) {
 			ftp_err(_("read error: %s\n"), strerror(errno));
 			ftp->ti.ioerror = true;
 		}
@@ -297,15 +417,37 @@ static int maybe_abort(FILE *in, FILE *out)
 			ftp_err(_("write error: %s\n"), strerror(errno));
 			ftp->ti.ioerror = true;
 		}
-		return ftp_abort(ftp->ti.transfer_is_put ? out : in);
+		return ftp_abort(in);
 	}
 	return 0;
 }
 
-static int FILE_recv_binary(FILE *in, FILE *out)
+static int maybe_abort_out(FILE *in, Socket *out)
+{
+	unsigned int i = ftp_sigints();
+	ftp_set_close_handler();
+
+	ftp->ti.finished = true;
+
+	if(ftp->ti.interrupted)
+		i++;
+
+	if(i > 0 || ferror(in) || sock_error_out(out)) {
+		if(ferror(in)) {
+			ftp_err(_("read error: %s\n"), strerror(errno));
+			ftp->ti.ioerror = true;
+		}
+		else if (sock_error_out(out)) {
+			ftp_err(_("write error: %s\n"), strerror(errno));
+			ftp->ti.ioerror = true;
+		}
+		return ftp_abort(out);
+	}
+	return 0;
+}
+
+static int FILE_recv_binary(Socket* in, FILE *out)
 {
-	size_t n;
-	char *buf;
 	time_t then = time(0) - 1;
 	time_t now;
 
@@ -315,22 +457,18 @@ static int FILE_recv_binary(FILE *in, FILE *out)
 		foo_hookf(&ftp->ti);
 	ftp->ti.begin = false;
 
-	clearerr(in);
+	sock_clearerr_in(in);
 	clearerr(out);
 
-	buf = (char *)xmalloc(FTP_BUFSIZ);
-	while(!feof(in)) {
-
+	char* buf = xmalloc(FTP_BUFSIZ);
+	while (!sock_eof(in)) {
 		if(wait_for_input() != 0) {
 			ftp_trace("wait_for_input() returned non-zero\n");
 			break;
 		}
-#ifdef SECFTP
-		n = sec_read(fileno(in), buf, FTP_BUFSIZ);
-#else
-		n = fread(buf, sizeof(char), FTP_BUFSIZ, in);
-#endif
-		if(n <= 0)
+
+    const ssize_t n = sock_read(in, buf, FTP_BUFSIZ);
+		if (n <= 0)
 			break;
 
 		if(ftp_sigints() > 0) {
@@ -355,13 +493,11 @@ static int FILE_recv_binary(FILE *in, FILE *out)
 	free(buf);
 	ftp_set_close_handler();
 
-	return maybe_abort(in, out);
+	return maybe_abort_in(in, out);
 }
 
-static int FILE_send_binary(FILE *in, FILE *out)
+static int FILE_send_binary(FILE *in, Socket *out)
 {
-	size_t n;
-	char *buf;
 	time_t then = time(0) - 1;
 	time_t now;
 
@@ -372,11 +508,11 @@ static int FILE_send_binary(FILE *in, FILE *out)
 	ftp->ti.begin = false;
 
 	clearerr(in);
-	clearerr(out);
+	sock_clearerr_out(out);
 
-	buf = (char *)xmalloc(FTP_BUFSIZ);
+	char* buf = xmalloc(FTP_BUFSIZ);
 	while(!feof(in)) {
-		n = fread(buf, sizeof(char), FTP_BUFSIZ, in);
+		ssize_t n = fread(buf, sizeof(char), FTP_BUFSIZ, in);
 		if(n <= 0)
 			break;
 
@@ -385,13 +521,9 @@ static int FILE_send_binary(FILE *in, FILE *out)
 
 		if(wait_for_output() != 0)
 			break;
-#ifdef SECFTP
-		if(sec_write(fileno(out), buf, n) != n)
-			break;
-#else
-		if(fwrite(buf, sizeof(char), n, out) != n)
-			break;
-#endif
+
+    if (sock_write(out, buf, n) != n)
+      break;
 		ftp->ti.size += n;
 		if(foo_hookf) {
 			now = time(0);
@@ -401,28 +533,14 @@ static int FILE_send_binary(FILE *in, FILE *out)
 			}
 		}
 	}
-#ifdef SECFTP
-	sec_fflush(out);
-#endif
-
+	sock_flush(out);
 	free(buf);
 
-	return maybe_abort(in, out);
+	return maybe_abort_out(in, out);
 }
 
-static int krb_getc(FILE *fp)
+static int FILE_recv_ascii(Socket* in, FILE *out)
 {
-#ifdef SECFTP
-	return sec_getc(fp);
-#else
-	return fgetc(fp);
-#endif
-}
-
-static int FILE_recv_ascii(FILE *in, FILE *out)
-{
-	char *buf = (char *)xmalloc(FTP_BUFSIZ);
-	int c;
 	time_t then = time(0) - 1;
 	time_t now;
 
@@ -432,28 +550,42 @@ static int FILE_recv_ascii(FILE *in, FILE *out)
 		foo_hookf(&ftp->ti);
 	ftp->ti.begin = false;
 
-	clearerr(in);
+	sock_clearerr_in(in);
 	clearerr(out);
 
-	while((c = krb_getc(in)) != EOF) {
-		if(ftp_sigints() > 0)
+  bool next = true;
+  int nc = 0;
+	while (true)
+  {
+    int c = 0;
+    if (next)
+    {
+      c = sock_get(in);
+      if (c == EOF)
+        break;
+    }
+    else
+    {
+      c = nc;
+      next = true;
+    }
+
+		if (ftp_sigints() > 0 || wait_for_input() != 0)
 			break;
 
-		if(wait_for_input() != 0)
-			break;
-
-		if(c == '\n')
+		if (c == '\n')
 			ftp->ti.barelfs++;
-		else if(c == '\r') {
-			c = krb_getc(in);
-			if(c == EOF)
+		else if (c == '\r')
+    {
+			nc = sock_get(in);
+			if (nc == EOF)
 				break;
-			if(c != '\n') {
-				ungetc(c, in);
-				c = '\r';
-			}
+			if (nc != '\n')
+        next = false;
+      else
+        c = nc;
 		}
-		if(fputc(c, out) == EOF)
+		if (fputc(c, out) == EOF)
 			break;
 
 		ftp->ti.size++;
@@ -466,15 +598,11 @@ static int FILE_recv_ascii(FILE *in, FILE *out)
 		}
 	}
 
-	free(buf);
-
-	return maybe_abort(in, out);
+	return maybe_abort_in(in, out);
 }
 
-static int FILE_send_ascii(FILE *in, FILE *out)
+static int FILE_send_ascii(FILE* in, Socket* out)
 {
-	char *buf = (char *)xmalloc(FTP_BUFSIZ);
-	int c;
 	time_t then = time(0) - 1;
 	time_t now;
 
@@ -485,8 +613,9 @@ static int FILE_send_ascii(FILE *in, FILE *out)
 	ftp->ti.begin = false;
 
 	clearerr(in);
-	clearerr(out);
+	sock_clearerr_out(out);
 
+  int c;
 	while((c = fgetc(in)) != EOF) {
 		if(ftp_sigints() > 0)
 			break;
@@ -495,11 +624,11 @@ static int FILE_send_ascii(FILE *in, FILE *out)
 			break;
 
 		if(c == '\n') {
-			if(fputc('\r', out) == EOF)
+			if(sock_put(out, '\r') == EOF)
 				break;
 			ftp->ti.size++;
 		}
-		if(fputc(c, out) == EOF)
+		if(sock_put(out, c) == EOF)
 			break;
 		ftp->ti.size++;
 		if(foo_hookf) {
@@ -511,9 +640,7 @@ static int FILE_send_ascii(FILE *in, FILE *out)
 		}
 	}
 
-	free(buf);
-
-	return maybe_abort(in, out);
+	return maybe_abort_out(in, out);
 }
 
 void reset_transfer_info(void)
@@ -536,44 +663,48 @@ void reset_transfer_info(void)
 
 int ftp_list(const char *cmd, const char *param, FILE *fp)
 {
-	if(!cmd || !fp || !ftp_connected())
-		return -1;
+  if (!cmd || !fp || !ftp_connected())
+    return -1;
 
-	if(ftp->ssh_pid)
-		return ssh_list(cmd, param, fp);
-
-	reset_transfer_info();
-	foo_hookf = 0;
-
-#if 0 /* don't care about transfer type, binary should work well... */
-	ftp_type(tmAscii);
+#ifdef HAVE_LIBSSH
+  if (ftp->session)
+    return ssh_list(cmd, param, fp);
 #endif
 
-	if(ftp_init_transfer() != 0)
-		return -1;
+  reset_transfer_info();
+  foo_hookf = NULL;
 
-	ftp_set_tmp_verbosity(vbNone);
-	if(param)
-		ftp_cmd("%s %s", cmd, param);
-	else
-		ftp_cmd("%s", cmd);
-	if(ftp->code != ctPrelim)
-		return -1;
+#if 0 /* don't care about transfer type, binary should work well... */
+  ftp_type(tmAscii);
+#endif
 
-	if(sock_accept(ftp->data, "r", ftp_is_passive()) != 0) {
-		perror("accept()");
-		return -1;
-	}
+  if (ftp_init_transfer() != 0) {
+    ftp_err(_("transfer initialization failed"));
+    return -1;
+  }
 
-	if(FILE_recv_ascii(ftp->data->sin, fp) != 0)
-		return -1;
+  ftp_set_tmp_verbosity(vbNone);
+  if (param)
+    ftp_cmd("%s %s", cmd, param);
+  else
+    ftp_cmd("%s", cmd);
+  if (ftp->code != ctPrelim)
+    return -1;
 
-	sock_destroy(ftp->data);
+  if (!sock_accept(ftp->data, "r", ftp_is_passive())) {
+    perror("accept()");
+    return -1;
+  }
 
-	ftp_read_reply();
+  if (FILE_recv_ascii(ftp->data, fp) != 0)
+    return -1;
 
-	ftp->data = 0;
-	return ftp->code == ctComplete ? 0 : -1;
+  sock_destroy(ftp->data);
+  ftp->data = NULL;
+
+  ftp_read_reply();
+
+  return ftp->code == ctComplete ? 0 : -1;
 }
 
 void transfer_finished(void)
@@ -612,7 +743,7 @@ static int ftp_init_receive(const char *path, transfer_mode_t mode,
 	if(ftp->code != ctPrelim)
 		return -1;
 
-	if(sock_accept(ftp->data, "r", ftp_is_passive()) != 0) {
+	if(!sock_accept(ftp->data, "r", ftp_is_passive())) {
 		ftp_err(_("data connection not accepted\n"));
 		return -1;
 	}
@@ -647,9 +778,9 @@ static int ftp_do_receive(FILE *fp,
 	int r;
 
 	if(mode == tmBinary)
-		r = FILE_recv_binary(ftp->data->sin, fp);
+		r = FILE_recv_binary(ftp->data, fp);
 	else
-		r = FILE_recv_ascii(ftp->data->sin, fp);
+		r = FILE_recv_ascii(ftp->data, fp);
 
 	sock_destroy(ftp->data);
 	ftp->data = 0;
@@ -671,8 +802,10 @@ static int ftp_do_receive(FILE *fp,
 int ftp_receive(const char *path, FILE *fp,
 				transfer_mode_t mode, ftp_transfer_func hookf)
 {
-	if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+	if(ftp->session)
 		return ssh_do_receive(path, fp, mode, hookf);
+#endif
 
 	if(ftp_init_receive(path, mode, hookf) != 0)
 		return -1;
@@ -690,6 +823,9 @@ static int ftp_send(const char *path, FILE *fp, putmode_t how,
 	if(how == putUnique && !ftp->has_stou_command)
 		return -1;
 
+  if (how == putTryUnique && !ftp->has_stou_command)
+    how = putNormal;
+
 	reset_transfer_info();
 	ftp->ti.transfer_is_put = true;
 
@@ -707,21 +843,29 @@ static int ftp_send(const char *path, FILE *fp, putmode_t how,
 		ftp->ti.restart_size = rp;
 	}
 
-	ftp_set_tmp_verbosity(vbError);
-	switch(how) {
-	  case putUnique:
-		ftp_cmd("STOU %s", path);
-		if(ftp->fullcode == 502)
-			ftp->has_stou_command = false;
-		break;
-	  case putAppend:
-		ftp_cmd("APPE %s", path);
-		break;
-	  case putNormal:
-	  default:
-		ftp_cmd("STOR %s", path);
-		break;
-	}
+  ftp_set_tmp_verbosity(vbError);
+  switch (how) {
+  case putAppend:
+    ftp_cmd("APPE %s", path);
+    break;
+
+  case putTryUnique:
+  case putUnique:
+    ftp_cmd("STOU %s", path);
+    if (ftp->fullcode == 502 || ftp->fullcode == 504) {
+      ftp->has_stou_command = false;
+      if (how == putTryUnique)
+        how = putNormal;
+      else
+        break;
+    }
+    else
+      break;
+
+  default:
+    ftp_cmd("STOR %s", path);
+    break;
+  }
 
 	if(ftp->code != ctPrelim)
 		return -1;
@@ -745,7 +889,7 @@ static int ftp_send(const char *path, FILE *fp, putmode_t how,
 		}
 	}
 
-	if(sock_accept(ftp->data, "w", ftp_is_passive()) != 0) {
+	if(!sock_accept(ftp->data, "w", ftp_is_passive())) {
 		ftp_err(_("data connection not accepted\n"));
 		return -1;
 	}
@@ -753,9 +897,9 @@ static int ftp_send(const char *path, FILE *fp, putmode_t how,
 	ftp_cache_flush_mark_for(path);
 
 	if(mode == tmBinary)
-		r = FILE_send_binary(fp, ftp->data->sin);
+		r = FILE_send_binary(fp, ftp->data);
 	else
-		r = FILE_send_ascii(fp, ftp->data->sin);
+		r = FILE_send_ascii(fp, ftp->data);
 	sock_flush(ftp->data);
 	sock_destroy(ftp->data);
 	ftp->data = 0;
@@ -794,17 +938,20 @@ int ftp_fxpfile(Ftp *srcftp, const char *srcfile,
 		return -1;
 	}
 
-	if(ftp->ssh_pid) {
+#ifdef HAVE_LIBSSH
+	if(ftp->session) {
 		ftp_err("FxP with SSH not implemented\n");
 		return -1;
 	}
+#endif
 
 	thisftp = ftp; /* save currently active connection */
 
 	/* setup source side */
 	ftp_use(srcftp);
 	ftp_type(mode);
-	if(ftp_pasv(addr) != 0) {
+  // TODO: IPv6 support
+	if(!ftp_pasv(false, addr, NULL)) {
 		ftp_use(thisftp);
 		return -1;
 	}
@@ -886,7 +1033,7 @@ int ftp_fxpfile(Ftp *srcftp, const char *srcfile,
 	ftp_cmd("RETR %s", srcfile);
 	if(ftp->code != ctPrelim) {
 		ftp_use(destftp);
-		ftp_abort(0);
+		ftp_abort(NULL);
 		ftp_use(thisftp);
 		return -1;
 	}
@@ -939,21 +1086,27 @@ int ftp_getfile(const char *infile, const char *outfile, getmode_t how,
 	rp = ftp->restart_offset;
 
 	reset_transfer_info();
-	if(ftp->ssh_pid) {
+#ifdef HAVE_LIBSSH
+	if (ftp->session)
+	{
 		/* we need to stat the remote file, so we are sure we can read it
 		 * this needs to be done before we call ssh_do_receive, because by
 		 * then, the file is created, and would leave a zero-size file opon
 		 * failure
 		 */
-		Attrib *a = ssh_stat(infile);
-		if(a == 0) {
-			ftp_err(_("Unable to stat file '%s'\n"), infile);
+		sftp_attributes a = sftp_stat(ftp->sftp_session, infile);
+		if(!a) {
+			ftp_err(_("Unable to stat file '%s': %s\n"), infile, ssh_get_error(ftp->session));
 			return -1;
 		}
+		sftp_attributes_free(a);
 		/* FIXME: how can we check if it will be possible to transfer
 		 * the specified file?
 		 */
-	} else if(ftp_init_receive(infile, mode, hookf) != 0)
+	}
+	else
+#endif
+	if (ftp_init_receive(infile, mode, hookf) != 0)
 		return -1;
 
 	if(how == getPipe) {
@@ -971,7 +1124,7 @@ int ftp_getfile(const char *infile, const char *outfile, getmode_t how,
 	}
 
 	if(rp > 0L) {
-		if(fseeko(fp, rp, SEEK_SET) != 0) {
+		if(fseek(fp, rp, SEEK_SET) != 0) {
 			ftp_err(_("%s: %s, transfer cancelled\n"),
 					outfile, strerror(errno));
 			close_func(fp);
@@ -987,9 +1140,11 @@ int ftp_getfile(const char *infile, const char *outfile, getmode_t how,
 
 	foo_hookf = hookf;
 
-	if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+	if(ftp->session)
 		r = ssh_do_receive(infile, fp, mode, hookf);
 	else
+#endif
 		r = ftp_do_receive(fp, mode, hookf);
 	close_func(fp);
 	return r;
@@ -1044,7 +1199,7 @@ int ftp_putfile(const char *infile, const char *outfile, putmode_t how,
 
 
 	if(ftp->restart_offset > 0L) {
-		if(fseeko(fp, ftp->restart_offset, SEEK_SET) != 0) {
+		if(fseek(fp, ftp->restart_offset, SEEK_SET) != 0) {
 			ftp_err(_("%s: %s, transfer cancelled\n"),
 					outfile, strerror(errno));
 			fclose(fp);
@@ -1054,9 +1209,11 @@ int ftp_putfile(const char *infile, const char *outfile, putmode_t how,
 
 	foo_hookf = hookf;
 
-	if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+	if(ftp->session)
 		r = ssh_send(outfile, fp, how, mode, hookf);
 	else
+#endif
 		r = ftp_send(outfile, fp, how, mode, hookf);
 	fclose(fp);
 	return r;
