@@ -1,5 +1,4 @@
-/* $Id: ftpsigs.c,v 1.5 2001/05/28 09:53:44 mhe Exp $
- *
+/*
  * ftpsigs.c -- handles signals
  *
  * Yet Another FTP Client
@@ -16,9 +15,7 @@
 #include "ftpsigs.h"
 #include "gvars.h"
 #include "transfer.h"
-
-/* in cmd.c */
-void exit_yafc(void);
+#include "cmd.h"
 
 void ftp_set_signal_with_mask(int signum, sighandler_t handler, int onesig)
 {
@@ -71,7 +68,7 @@ void ftp_set_signal(int signum, sighandler_t handler)
 
 static unsigned int sigints = 0;
 
-RETSIGTYPE sigint_close_handler(int signum)
+void sigint_close_handler(int signum)
 {
 	if(gvSighupReceived)
 		return;
@@ -91,7 +88,11 @@ RETSIGTYPE sigint_close_handler(int signum)
 	 * program?
 	 */
 
-	if(sigints >= 4 || ftp->ssh_pid) {
+	if (sigints >= 4
+#ifdef HAVE_LIBSSH
+			|| ftp->session
+#endif
+		) {
 		ftp_close();
 		sigints = 0;
 		if(gvJmpBufSet) {
@@ -116,7 +117,7 @@ RETSIGTYPE sigint_close_handler(int signum)
 /*	ftp_set_signal(SIGINT, sigint_close_handler);*/
 }
 
-RETSIGTYPE sigint_abort_handler(int signum)
+void sigint_abort_handler(int signum)
 {
 	if(gvSighupReceived)
 		return;
@@ -126,7 +127,11 @@ RETSIGTYPE sigint_abort_handler(int signum)
 
 	gvInterrupted = true;
 
-	if(sigints >= 3 || ftp->ssh_pid) {
+	if (sigints >= 3
+#ifdef HAVE_LIBSSH
+			|| ftp->session
+#endif
+		) {
 		sigints = 0;
 		alarm(0);
 		if(gvJmpBufSet) {
@@ -142,7 +147,7 @@ RETSIGTYPE sigint_abort_handler(int signum)
 /*	ftp_set_signal(SIGINT, sigint_abort_handler);*/
 }
 
-RETSIGTYPE sigint_jmp_handler(int signum)
+void sigint_jmp_handler(int signum)
 {
 	if(gvSighupReceived)
 		return;
@@ -151,8 +156,10 @@ RETSIGTYPE sigint_jmp_handler(int signum)
 	alarm(0);
 	if(gvJmpBufSet) {
 		ftp_trace("jumping to gvRestartJmp\n");
-		if(ftp->ssh_pid)
+#ifdef HAVE_LIBSSH
+		if(ftp->session)
 			ftp_close();
+#endif
 		ftp_longjmp(gvRestartJmp, 1);
 	} else {
 		ftp_err(_("Interrupt received, exiting...         \n"));
@@ -161,7 +168,7 @@ RETSIGTYPE sigint_jmp_handler(int signum)
 }
 
 #if 0
-static RETSIGTYPE sighup_term_handler(int signum)
+static void sighup_term_handler(int signum)
 {
 	printf(_("SIGTERM (terminate) received, exiting...\n"));
 	ftp_close();
@@ -171,7 +178,7 @@ static RETSIGTYPE sighup_term_handler(int signum)
 }
 #endif
 
-RETSIGTYPE sighup_handler(int signum)
+void sighup_handler(int signum)
 {
 	ftp_set_signal(SIGINT, SIG_IGN);
 	ftp_set_signal(SIGHUP, sighup_handler);
